@@ -8,7 +8,7 @@ from sortedcontainers import SortedDict
 import elftools
 from elftools.elf import elffile, sections, dynamic, enums
 from elftools.dwarf import callframe
-from elftools.common.exceptions import ELFError, DWARFError
+from elftools.common.exceptions import ELFError, DWARFError, ELFParseError
 from elftools.dwarf.dwarf_expr import DWARFExprParser
 from elftools.dwarf.descriptions import describe_form_class, describe_attr_value
 from elftools.dwarf.dwarfinfo import DWARFInfo
@@ -19,7 +19,7 @@ import archinfo
 from .compilation_unit import CompilationUnit
 from .variable import Variable
 from .variable_type import VariableType
-from .subprogram import Subprogram
+from .subprogram import LexicalBlock, Subprogram
 from .symbol import ELFSymbol, Symbol, SymbolType
 from .regions import ELFSection, ELFSegment
 from .hashtable import ELFHashTable, GNUHashTable
@@ -43,17 +43,33 @@ __all__ = ('ELFSymbol', 'ELF')
 additional_e_machine_mappings : Dict[int, str] = {
     247: 'Linux BPF - in-kernel virtual machine',
     252: 'C-SKY',
-    0x5441: 'Fujitsu FR-V'
+    0x5441: 'Fujitsu FR-V',
+    0x9026: 'DEC Alpha',
 }
 
 
 class ELF(MetaELF):
     """
     The main loader class for statically loading ELF executables. Uses the pyreadelf library where useful.
+
+    Useful backend options:
+
+    - ``debug_symbols``: Provides the path to a separate file which contains the binary's debug symbols
+    - ``discard_section_headers``: Do not parse section headers. Use this if they are corrupted or malicious.
+    - ``discard_program_headers``: Do not parse program headers. Use this if the binary is for a platform whose ELF
+            loader only looks at section headers, but whose toolchain generates program headers anyway.
     """
     is_default = True  # Tell CLE to automatically consider using the ELF backend
 
-    def __init__(self, *args, addend=None, debug_symbols=None, **kwargs):
+    def __init__(
+            self,
+            *args,
+            addend=None,
+            debug_symbols=None,
+            discard_section_headers=False,
+            discard_program_headers=False,
+            **kwargs
+    ):
         super().__init__(*args, **kwargs)
         patch_undo = []
         try:
@@ -147,8 +163,10 @@ class ELF(MetaELF):
         except ValueError:
             l.info('no PT_LOAD segments identified')
 
-        self.__register_segments()
-        self.__register_sections()
+        if not discard_program_headers:
+            self.__register_segments()
+        if not discard_section_headers:
+            self.__register_sections()
 
         if not self.symbols:
             self._desperate_for_symbols = True
@@ -177,12 +195,13 @@ class ELF(MetaELF):
             self.__process_debug_file(debug_symbols)
         elif self.loader._load_debug_info:
             if self.build_id:
-                debug_filename = '/usr/lib/debug/.build-id/%s/%s.debug' % (self.build_id[:2], self.build_id[2:])
+                debug_filename = f'/usr/lib/debug/.build-id/{self.build_id[:2]}/{self.build_id[2:]}.debug'
+                if os.path.isfile(debug_filename):
+                    self.__process_debug_file(debug_filename)
+            if self.binary:
+                debug_filename = os.path.join('/usr/lib/debug', os.path.realpath(self.binary))
                 if os.path.isfile(debug_filename):
                     self.__process_debug_file(debug_filename)
-            debug_filename = os.path.join('/usr/lib/debug', os.path.realpath(self.binary))
-            if os.path.isfile(debug_filename):
-                self.__process_debug_file(debug_filename)
 
         # call the methods defined by MetaELF
         self._ppc64_abiv1_entry_fix()
@@ -371,7 +390,7 @@ class ELF(MetaELF):
                 return cached
             re_sym = symid
         else:
-            raise CLEError("Bad symbol identifier: %r" % (symid,))
+            raise CLEError(f"Bad symbol identifier: {symid!r}")
 
         symbol = ELFSymbol(self, re_sym)
         if version is not None and self._versions is not None:
@@ -575,10 +594,17 @@ class ELF(MetaELF):
         """
         for cu in dwarf.iter_CUs():
             comp_dir = '.'
-            die = cu.get_top_DIE()
+            try:
+                die = cu.get_top_DIE()
+            except KeyError:
+                # pyelftools is not very resilient
+                continue
             if 'DW_AT_comp_dir' in die.attributes:
                 comp_dir = die.attributes['DW_AT_comp_dir'].value.decode()
-            lineprog = dwarf.line_program_for_CU(cu)
+            try:
+                lineprog = dwarf.line_program_for_CU(cu)
+            except ELFParseError:
+                continue
             if lineprog is None:
                 continue
             file_cache = {}
@@ -646,11 +672,15 @@ class ELF(MetaELF):
             expr_parser = DWARFExprParser(cu.structs)
 
             # scan the whole die tree for DW_TAG_base_type
-            for die in cu.iter_DIEs():
-                if die.tag == "DW_TAG_base_type":
-                    var_type = VariableType.read_from_die(die)
-                    if var_type is not None:
-                        type_list[die.offset] = var_type
+            try:
+                for die in cu.iter_DIEs():
+                    if VariableType.supported_die(die):
+                        var_type = VariableType.read_from_die(die, self)
+                        if var_type is not None:
+                            type_list[die.offset] = var_type
+            except KeyError:
+                # pyelftools is not very resilient - we need to catch KeyErrors here
+                continue
 
             top_die = cu.get_top_DIE()
 
@@ -671,77 +701,52 @@ class ELF(MetaELF):
             die_comp_dir = die_comp_dir.value.decode('utf-8')
             die_lang = describe_attr_value(die_lang, top_die, top_die.offset)
 
-            cu_ = CompilationUnit(die_name, die_comp_dir, die_low_pc, die_high_pc,die_lang)
+            cu_ = CompilationUnit(die_name, die_comp_dir, die_low_pc, die_high_pc,die_lang, self)
             compilation_units.append(cu_)
 
             for die_child in cu.iter_DIE_children(top_die):
                 if die_child.tag == 'DW_TAG_variable':
                     # load global variable
-                    var = self._load_die_variable(die_child, expr_parser, type_list)
+                    var = Variable.from_die(die_child, expr_parser, self)
                     var.decl_file = cu_.file_path
                     cu_.global_variables.append(var)
                 elif die_child.tag == 'DW_TAG_subprogram':
                     # load subprogram
+                    sub_prog = self._load_die_lex_block(die_child, expr_parser, type_list, cu, cu_.file_path, None)
+                    if sub_prog is not None:
+                        cu_.functions[sub_prog.low_pc] = sub_prog
 
-                    if 'DW_AT_name' in die_child.attributes:
-                        name = die_child.attributes['DW_AT_name'].value.decode('utf-8')
-                    else:
-                        name = None
-                    low_pc, high_pc = self._load_low_high_pc_form_die(die_child)
-                    if low_pc is not None or high_pc is not None:
-                        sub_prog = Subprogram(name, low_pc, high_pc)
-
-                        for sub_die in cu._iter_DIE_subtree(die_child):
-                            if sub_die.tag in ['DW_TAG_variable','DW_TAG_formal_parameter']:
-                                # load local variable
-                                var = self._load_die_variable(sub_die, expr_parser, type_list)
-                                var.decl_file = cu_.file_path
-                                sub_prog.local_variables.append(var)
-
-                        cu_.functions[low_pc] = sub_prog
-
+        self.type_list = type_list
         self.compilation_units = compilation_units
 
-    @staticmethod
-    def _load_die_variable(die: DIE, expr_parser, type_list) -> Variable:
+    def _load_die_lex_block(self, die: DIE, expr_parser, type_list, cu, file_path, subprogram) -> LexicalBlock:
 
         if 'DW_AT_name' in die.attributes:
-            var_name = die.attributes['DW_AT_name'].value.decode('utf-8')
+            name = die.attributes['DW_AT_name'].value.decode('utf-8')
         else:
-            var_name = None
+            name = None
 
-        if 'DW_AT_decl_line' in die.attributes:
-            decl_line = die.attributes['DW_AT_decl_line'].value
+        low_pc, high_pc = self._load_low_high_pc_form_die(die)
+        if low_pc is None or high_pc is None:
+            return None
+
+        if subprogram is None:
+            subprogram = block = Subprogram(name, low_pc, high_pc)
         else:
-            decl_line = None
+            block = LexicalBlock(low_pc, high_pc)
 
-        type_ = None
-        if 'DW_AT_type' in die.attributes:
-            type_offset = die.attributes['DW_AT_type'].value
-            if type_offset in type_list:
-                type_ = type_list[type_offset]
-            else:
-                # this warning takes too long to print on binaries with too many unknown type offsets. disable it
-                # l.warning("unknown type offset var_name:%s type_offset:%s", var_name, type_offset)
-                pass
+        for sub_die in cu.iter_DIE_children(die):
+            if sub_die.tag in ['DW_TAG_variable', 'DW_TAG_formal_parameter']:
+                # load local variable
+                var = Variable.from_die(sub_die, expr_parser, self, block)
+                var.decl_file = file_path
+                subprogram.local_variables.append(var)
+            elif sub_die.tag == 'DW_TAG_lexical_block':
+                sub_block = self._load_die_lex_block(sub_die, expr_parser, type_list, cu, file_path, subprogram)
+                if sub_block is not None:
+                    block.child_blocks.append(sub_block)
 
-        v = Variable(
-            name= var_name,
-            type_= type_,
-            decl_file= None,  # decl_file: will be back-patched in _load_dies()
-            decl_line= decl_line,
-        )
-
-        if 'DW_AT_location' in die.attributes and die.attributes['DW_AT_location'].form == 'DW_FORM_exprloc':
-            parsed_exprs = expr_parser.parse_expr(die.attributes['DW_AT_location'].value)
-            if len(parsed_exprs) == 1 and parsed_exprs[0].op_name == 'DW_OP_addr':
-                v.sort, v.addr = "global", parsed_exprs[0].args[0]
-            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name == 'DW_OP_fbreg':
-                v.sort, v.addr = "stack", parsed_exprs[0].args[0]
-            elif len(parsed_exprs) == 1 and parsed_exprs[0].op_name.startswith("DW_OP_reg"):
-                v.sort, v.addr = "register", parsed_exprs[0].op - 0x50 # 0x50 == DW_OP_reg0
-
-        return v
+        return block
 
     #
     # Private Methods... really. Calling these out of context
@@ -762,7 +767,8 @@ class ELF(MetaELF):
         for seg in type_to_seg_mapping['PT_LOAD']:
             self._load_segment(seg)
 
-        # the order of processing for the other three handled segment_types should not matter, but let's have it consistent
+        # the order of processing for the other three handled segment_types should not matter, but let's have
+        # it consistent
         for seg in type_to_seg_mapping['PT_DYNAMIC']:
             self.__register_dyn(seg)
             self.linking = 'dynamic'
@@ -928,7 +934,7 @@ class ELF(MetaELF):
             if rela_tag in self._dynamic:
                 reloffset = AT.from_lva(self._dynamic[rela_tag], self).to_rva()
                 if relsz_tag not in self._dynamic:
-                    raise CLEInvalidBinaryError('Dynamic section contains %s but not %s' % (rela_tag, relsz_tag))
+                    raise CLEInvalidBinaryError(f'Dynamic section contains {rela_tag} but not {relsz_tag}')
                 relsz = self._dynamic[relsz_tag]
                 fakerelheader = {
                     'sh_offset': reloffset,
@@ -1231,8 +1237,8 @@ class ELF(MetaELF):
 
     def __neuter_streams(self, obj):
         if isinstance(obj, elftools.elf.dynamic._DynamicStringTable):
-           obj._stream = self.memory
-           obj._table_offset = self._offset_to_rva(obj._table_offset)
+            obj._stream = self.memory
+            obj._table_offset = self._offset_to_rva(obj._table_offset)
         elif isinstance(obj, elftools.elf.sections.Section):
             if obj.header.sh_type == 'SHT_NOBITS':
                 obj.stream = None
