@@ -1,22 +1,25 @@
-# -*-coding:utf8 -*-
 # This file is part of Mach-O Loader for CLE.
 # Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/) and updated in September 2019.
 
 import struct
-from typing import Callable, Dict, Tuple
+from typing import Callable, Dict, Tuple, TYPE_CHECKING
+import logging
+
+
+from ..relocation import Relocation
+from .. import Backend
+
+from .symbol import BindingSymbol, AbstractMachOSymbol, SymbolTableSymbol, DyldBoundSymbol
 
-from .symbol import BindingSymbol
 
-from typing import TYPE_CHECKING
 
-if TYPE_CHECKING:
-    from ... import MachO
 
 from ...errors import CLEInvalidBinaryError
 from ...address_translator import AT
 
-import logging
-logging.basicConfig(level=logging.INFO)
+if TYPE_CHECKING:
+    from ... import MachO
+
 l = logging.getLogger(name=__name__)
 
 OPCODE_MASK = 0xF0
@@ -95,7 +98,8 @@ class BindingState():
         self.segment_index = 0
         self.address = 0
         self.seg_end_address = 0  # TODO: no rebasing support
-        self.wraparound = 2 ** 64  # address is expected to properly overflow and address is uintptr_t (unsigned long according to _uintptr_t.h)
+        # address is expected to properly overflow and address is uintptr_t (unsigned long according to _uintptr_t.h)
+        self.wraparound = 2 ** 64
         self.sizeof_intptr_t = 8 if is_64 else 4  # experimentally determined
         self.bind_handler = None  # function(state,binary) => None
 
@@ -367,6 +371,42 @@ def n_opcode_do_bind_uleb_times_skipping_uleb(s: BindingState, b: 'MachO', _i: i
     return s
 
 
+class MachORelocation(Relocation):
+    """
+    Generic Relocation for MachO. For now it just deals with symbols
+    """
+    def __init__(self, owner: Backend, symbol: AbstractMachOSymbol, relative_addr: int, data):
+        super().__init__(owner, symbol, relative_addr)
+        self.data = data
+
+    def resolve_symbol(self, solist, thumb=False, extern_object=None, **kwargs):
+        if isinstance(self.symbol, (SymbolTableSymbol, BindingSymbol, DyldBoundSymbol)):
+            if self.symbol.library_name in [so.binary_basename for so in solist]:
+                raise NotImplementedError(
+                    "Symbol could actually be resolved because we have the required library,"
+                    " but that isn't implemented yet")
+            else:
+                # Create an extern symbol for it
+                new_symbol = extern_object.make_extern(self.symbol.name, sym_type=self.symbol._type, thumb=thumb)
+                self.resolve(new_symbol, extern_object=extern_object)
+        else:
+            raise NotImplementedError("Did not expect this to happen")
+
+    @property
+    def dest_addr(self):
+        """
+        mach-o rebasing is hard to handle, so this behaviour differs from other relocations
+        """
+        return self.rebased_addr
+
+
+    @property
+    def value(self):
+        return self.resolvedby.rebased_addr
+
+    def __repr__(self):
+        return f"<MachO Reloc for {self.symbol} at {hex(self.relative_addr)}>"
+
 # default binding handler
 def default_binding_handler(state: BindingState, binary: 'MachO'):
     """Binds location to the symbol with the given name and library ordinal
@@ -393,9 +433,10 @@ def default_binding_handler(state: BindingState, binary: 'MachO'):
 
     if state.binding_type == 1:  # POINTER
         l.debug("Updating address %#x with symobl %r @ %#x", location, state.sym_name, value)
-        binary.memory.store(
-            AT.from_lva(location, binary).to_rva(),
-            struct.pack(binary.struct_byteorder + ("Q" if binary.arch.bits == 64 else "I"), value))
+        addr = AT.from_mva(location, binary).to_rva()
+        data = struct.pack(binary.struct_byteorder + ("Q" if binary.arch.bits == 64 else "I"), value)
+        reloc = MachORelocation(binary, symbol, addr, data)
+        binary.relocs.append(reloc)
         symbol.bind_xrefs.append(location)
     elif state.binding_type == 2:  # ABSOLUTE32
         location_32 = location % (2 ** 32)
