@@ -3,7 +3,7 @@ import sys
 import platform
 import logging
 from collections import OrderedDict
-from typing import Optional
+from typing import Optional, List
 
 import archinfo
 from archinfo.arch_soot import ArchSoot
@@ -11,11 +11,6 @@ from archinfo.arch_soot import ArchSoot
 from .address_translator import AT
 from .utils import ALIGN_UP, key_bisect_floor_key, key_bisect_insort_right
 
-try:
-    import claripy
-except ImportError:
-    claripy = None
-
 __all__ = ('Loader',)
 
 l = logging.getLogger(name=__name__)
@@ -51,8 +46,8 @@ class Loader:
     :param rebase_granularity:  The alignment to use for rebasing shared objects
     :param except_missing_libs: Throw an exception when a shared library can't be found.
     :param aslr:                Load libraries in symbolic address space. Do not use this option.
-    :param page_size:           The granularity with which data is mapped into memory. Set to 1 if you are working
-                                in a non-paged environment.
+    :param page_size:           The granularity with which data is mapped into memory. Set to 0x1000 if you are working
+                                in an environment where data will always be memory mapped in a page-graunlar way.
     :param preload_libs:        Similar to `force_load_libs` but will provide for symbol resolution, with precedence
                                 over any dependencies.
     :ivar memory:               The loaded, rebased, and relocated memory of the program.
@@ -100,7 +95,7 @@ class Loader:
 
         self._auto_load_libs = auto_load_libs
         self._load_debug_info = load_debug_info
-        self._satisfied_deps = dict((x, False) for x in skip_libs)
+        self._satisfied_deps = {x: False for x in skip_libs}
         self._main_opts = {} if main_opts is None else main_opts
         self._lib_opts = {} if lib_opts is None else lib_opts
         self._custom_ld_path = [ld_path] if type(ld_path) is str else ld_path
@@ -130,7 +125,7 @@ class Loader:
         self._kernel_object = None # type: Optional[KernelObject]
         self._extern_object = None # type: Optional[ExternObject]
         self.shared_objects = OrderedDict()
-        self.all_objects = []  # this list should always be sorted by min_addr
+        self.all_objects = []  # type: List[Backend]
         self.requested_names = set()
         if arch is not None:
             self._main_opts.update({'arch': arch})
@@ -150,9 +145,9 @@ class Loader:
 
     def __repr__(self):
         if self._main_binary_stream is None:
-            return '<Loaded %s, maps [%#x:%#x]>' % (os.path.basename(self._main_binary_path), self.min_addr, self.max_addr)
+            return f'<Loaded {os.path.basename(self._main_binary_path)}, maps [{self.min_addr:#x}:{self.max_addr:#x}]>'
         else:
-            return '<Loaded from stream, maps [%#x:%#x]>' % (self.min_addr, self.max_addr)
+            return f'<Loaded from stream, maps [{self.min_addr:#x}:{self.max_addr:#x}]>'
 
     @property
     def max_addr(self):
@@ -196,6 +191,16 @@ class Loader:
                 return obj
         return None
 
+    @property
+    def elfcore_object(self):
+        """
+        If a corefile was loaded, this returns the actual core object instead of the main binary
+        """
+        for obj in self.all_objects:
+            if isinstance(obj, ELFCore):
+                return obj
+        return None
+
     @property
     def extern_object(self):
         """
@@ -254,13 +259,13 @@ class Loader:
         """
         Return a set of every name that was requested as a shared object dependency but could not be loaded
         """
-        return self.requested_names - set(k for k,v in self._satisfied_deps.items() if v is not False)
+        return self.requested_names - {k for k,v in self._satisfied_deps.items() if v is not False}
 
     @property
     def auto_load_libs(self):
         return self._auto_load_libs
 
-    def describe_addr(self, addr):
+    def describe_addr(self, addr) -> str:
         """
         Returns a textual description of what's in memory at the provided address
         """
@@ -302,7 +307,7 @@ class Loader:
             objname = 'object loaded from stream'
 
         best_offset, best_prefix = max(options, key=lambda v: v[0])
-        return '%s%#x in %s (%#x)' % (best_prefix, rva - best_offset, objname, AT.from_va(addr, o).to_lva())
+        return f'{best_prefix}{rva - best_offset:#x} in {objname} ({AT.from_va(addr, o).to_lva():#x})'
 
     # Search functions
 
@@ -422,6 +427,27 @@ class Loader:
 
         return obj.find_section_containing(addr)
 
+    def find_loadable_containing(self, addr, skip_pseudo_objects=True):
+        """
+        Find the section or segment object the address belongs to. Sections will only be used if the corresponding
+        object does not have segments.
+
+        :param addr: The address to test
+        :param skip_pseudo_objects: Skip objects that CLE adds during loading.
+        :return:  The section or segment that the address belongs to, or None if the address does not belong to any
+                    section or segment.
+        """
+        obj = self.find_object_containing(addr, membership_check=False)
+
+        if obj is None:
+            return None
+
+        if skip_pseudo_objects and isinstance(obj, (ExternObject, KernelObject, TLSObject)):
+            # the address is from a special CLE section
+            return None
+
+        return obj.find_loadable_containing(addr)
+
     def find_section_next_to(self, addr, skip_pseudo_objects=True):
         """
         Find the next section after the given address.
@@ -527,17 +553,19 @@ class Loader:
         :param exclude_forwards:    Whether to exclude forward symbols. Default True.
         """
         for so in self.all_objects:
-            sym = so.get_symbol(name)
-            if sym is None:
-                continue
-            if sym.is_import and exclude_imports:
-                continue
-            if sym.owner is self._extern_object and exclude_externs:
-                continue
-            if sym.is_forward and exclude_forwards:
-                continue
+            for sym in so.symbols:
+                if sym is None:
+                    continue
+                if sym.name != name:
+                    continue
+                if sym.is_import and exclude_imports:
+                    continue
+                if sym.owner is self._extern_object and exclude_externs:
+                    continue
+                if sym.is_forward and exclude_forwards:
+                    continue
 
-            yield sym
+                yield sym
 
     def find_plt_stub_name(self, addr):
         """
@@ -600,6 +628,12 @@ class Loader:
         """
         if not self.aslr:
             return []
+
+        try:
+            import claripy  # pylint:disable=import-outside-toplevel
+        except ImportError:
+            claripy = None
+
         if not claripy:
             l.error("Please install claripy to get symbolic constraints")
             return []
@@ -843,7 +877,7 @@ class Loader:
             if close:
                 binary_stream.close()
 
-    def _map_object(self, obj):
+    def _map_object(self, obj: 'Backend'):
         """
         This will integrate the object into the global address space, but will not perform relocations.
         """
@@ -863,7 +897,7 @@ class Loader:
 
             obj.rebase(base_addr)
         else:
-            if obj._custom_base_addr is not None and not isinstance(obj, Blob):
+            if obj._custom_base_addr is not None and obj.linked_base != obj._custom_base_addr and not isinstance(obj, Blob):
                 l.warning("%s: base_addr was specified but the object is not PIC. "
                           "specify force_rebase=True to override", obj.binary_basename)
             base_addr = obj.linked_base
@@ -898,14 +932,23 @@ class Loader:
             else:
                 gap_start = ALIGN_UP(o.max_addr + 1, self._rebase_granularity)
 
-        if gap_start + size >= 2**self.main_object.arch.bits:
-            raise CLEOperationError("Ran out of room in address space")
+        if gap_start + size > 2**self.main_object.arch.bits:
+            # this may happen when loading an ELF core whose main object may occupy a large range of memory addresses
+            # with large unoccupied holes left in the middle
+            # we fall back to finding unoccupied holes
+            for this_seg, next_seg in zip(self.main_object.segments.raw_list, self.main_object.segments.raw_list[1:]):
+                gap_start = ALIGN_UP(this_seg.vaddr + this_seg.memsize, self._rebase_granularity)
+                gap = next_seg.vaddr - gap_start
+                if gap >= size:
+                    break
+            else:
+                raise CLEOperationError("Ran out of room in address space")
 
         return gap_start
 
     def _is_range_free(self, va, size):
         # self.main_object should not be None here
-        if va < 0 or va + size >= 2**self.main_object.arch.bits:
+        if va < 0 or va + size > 2**self.main_object.arch.bits:
             return False
 
         for o in self.all_objects:
@@ -968,6 +1011,9 @@ class Loader:
                 dirs.extend(self.main_object.extra_load_path)
                 if sys.platform.startswith('linux'):
                     dirs.extend(self.main_object.arch.library_search_path())
+                elif sys.platform.startswith('openbsd'):
+                    dirs.extend(self.main_object.arch.library_search_path())
+                    dirs.extend(['/usr/local/lib', '/usr/X11R6/lib'])
                 elif sys.platform == 'win32':
                     native_dirs = os.environ['PATH'].split(';')
 
@@ -1010,7 +1056,7 @@ class Loader:
                         ilibname = libname.lower() if self._case_insensitive else libname
                         if ilibname.strip('.0123456789') == spec.strip('.0123456789'):
                             yield os.path.realpath(os.path.join(libdir, libname))
-                except (IOError, OSError): pass
+                except OSError: pass
 
     @classmethod
     def _path_insensitive(cls, path):
